from __future__ import annotations

import abc

from typing import List, Iterable, Type

from ..model import GraphElement
from .lang import create_query, prepare_params
from ...exceptions import SQErzoElementExistException, SQErzoException
from ..interfaces import SQErzoGraphConnection, ResultElement


class CypherSQErzoGraphConnection(SQErzoGraphConnection):

    def plain_string(self, text: str) -> str:
        return text.replace("-", "_").replace(":", "_").lower()

    def labels_to_name(self, labels: List[str]) -> str:
        return "_".join(l.replace(":", "_") for l in labels)

    def get_node_by_id(self, node_id: str) \
            -> GraphElement or None:

        if not node_id:
            return

        q = f"""
        MATCH (n {{ identity: '{node_id}' }})
        RETURN n
        """
        if result := self.query_response(q):
            for res in result:
                if type(res) is list:
                    return res[0]

                else:
                    return res

    def save_element(self, graph_element: GraphElement) \
            -> None or SQErzoElementExistException:

        self.query(create_query(graph_element))

    def update_element(self, graph_element: GraphElement) \
            -> None or SQErzoElementExistException:

        self.query(graph_element.query_update())

        graph_element.__dirty_properties__.clear()

    def fetch_nodes(self,
                    node_type: Type[GraphElement] or GraphElement,
                    **kwargs) \
            -> Iterable[GraphElement] or SQErzoException:

        #
        # Detect if 'node_type' is and object instance or a class definition
        #
        if "__module__" in node_type.__dict__:
            # Is a class definition
            labels = ":".join(node_type.__labels__)
            class_constructor = node_type
        else:

            # Is an object instance
            if not issubclass(type(node_type), GraphElement):
                raise SQErzoException(
                    "'node_type' must be 'GraphElement' subclass"
                )

            labels = node_type.labels()
            class_constructor = node_type.__class__

        tmp_prop = " and ".join(
            prepare_params(kwargs, operation="insert")
        )

        query = f"""
        MATCH (a:{labels} {{ {tmp_prop} }})
        RETURN a
        """

        with self.query_response(query) as responses:

            for res in responses:

                yield_values = []
                for r in res:
                    yield_values.append(
                        class_constructor.from_query_results(r)
                    )
                # yield class_constructor.from_query_results(res)

                if len(yield_values) > 1:
                    yield yield_values
                else:
                    yield yield_values[0]

    @property
    def batch_size(self) -> int:
        return 1000

    @property
    def transaction_class(self) -> Type:
        from .transaction import CypherSQErzoTransaction

        return CypherSQErzoTransaction

    def truncate(self):
        self.query("match (p) detach delete p")

__all__ = ("CypherSQErzoGraphConnection", "ResultElement")



import urllib.parse as pr

from typing import List, Iterable, Callable

from neo4j import GraphDatabase

from ..model import GraphElement
from ..interfaces import ResultElement, SQErzoQueryResponse
from ... import SQErzoGraphConnection
from ...exceptions import SQErzoElementExistException
from .interfaces import CypherSQErzoGraphConnection


class Neo4jSQErzoQueryResponse(SQErzoQueryResponse):

    def __init__(self,
                 graph: SQErzoGraphConnection,
                 query: str,
                 **kwargs
                 ):
        self.query = query
        self.graph = graph
        self.params = kwargs

    def __iter__(self):
        with self.graph.connection.session() as session:

            ret = session.run(self.query, **self.params)

            for record in ret:

                res = []

                for k in record.keys():
                    for node in record.values(k):
                        res.append(ResultElement(
                            id=node.id,
                            properties=node._properties,
                            labels=list(node.labels),
                            alias=k
                        ))

                yield res

class Neo4JSQErzoGraphConnection(CypherSQErzoGraphConnection):

    SUPPORTED_TYPES = ("str", "int", "float", "bool", "datetime")

    def __init__(self, connection_string: str):
        self.connection = self._parse_connection_string(connection_string)
        self.enterprise: bool = False

    def query_with_response(self,
                            query: str,
                            **kwargs) -> Iterable[ResultElement]:
        with self.connection.session() as session:
            session.run(query, **kwargs)


    def query_response(self, query: str, **kwargs) -> Iterable[ResultElement]:
        return Neo4jSQErzoQueryResponse(
            self,
            query,
            **kwargs
        )

    def query(self, query: str, **kwargs) -> None or object:
        with self.connection.session() as session:
            return session.run(query, **kwargs)

    def save_element(self, graph_element: GraphElement) -> None or SQErzoElementExistException:
        try:
            super(Neo4JSQErzoGraphConnection, self).save_element(graph_element)
        except Exception as e:
            raise SQErzoElementExistException(
                f"Graph element with id '{graph_element.identity}' already exits"
            )

    def create_constraints_nodes(self, key: str, labels: List[str]):
        constrain_name = self.plain_string(self.labels_to_name(labels))
        label = self.labels_to_name(labels)

        q = f"""
        CREATE CONSTRAINT {constrain_name}_unique_{key} IF NOT EXISTS
        ON (p:{label}) ASSERT p.{key} IS UNIQUE
        """

        self.query(q)

    def create_constraints_edges(self, key: str, labels: List[str]):
        if not self.enterprise:
            return

        constrain_name = self.plain_string(self.labels_to_name(labels))
        label = self.labels_to_name(labels)

        q = f"""
        CREATE CONSTRAINT {constrain_name}_unique_{key} IF NOT EXISTS
        ON ()-[p:{label}]-() ASSERT EXISTS (p.{key})
        """

        self.query(q)

    def create_indexes(self, attribute: str, labels: List[str]):
        label = self.labels_to_name(labels)
        index_suffix = self.plain_string(label)
        index_prefix = self.plain_string(attribute)

        q = f"""
        CREATE INDEX index_{index_prefix}_{index_suffix} IF NOT EXISTS FOR (n:{label})
        ON (n.{attribute})
        """

        self.query(q)

    def _parse_connection_string(self, cs: str) -> GraphDatabase:
        parsed = pr.urlparse(cs)

        host = parsed.hostname
        if not host:
            host = "127.0.0.1"

        port = parsed.port
        if port is None:
            port = 7687

        config = {"uri": f"bolt://{host}:{port}"}

        if parsed.username and parsed.password:
            config["auth"] = (parsed.username, parsed.password)

        if "enterprise" in parsed.scheme:
            self.enterprise = True

        return GraphDatabase.driver(**config)


__all__ = ("Neo4JSQErzoGraphConnection",)



import redis
import urllib.parse as pr

from typing import Iterable, List, Type, Tuple

from redisgraph import Graph

from ...exceptions import *
from ..model import GraphElement, GraphNode
from .transaction import CypherSQErzoTransaction
from .interfaces import ResultElement, CypherSQErzoGraphConnection
from ..interfaces import SQErzoQueryResponse, SQErzoGraphConnection


class RedisGraphSQErzoQueryResponse(SQErzoQueryResponse):

    def __init__(self, graph: SQErzoGraphConnection, query: str, **kwargs):
        self.query = query
        self.graph = graph
        self.params = kwargs

    def __iter__(self):
        query_results = self.graph.connection.query(self.query, self.params)

        for res in query_results.result_set:

            yield [
                    ResultElement(
                        id=node.id,
                        alias=query_results.header[i][1].decode(),
                        labels=list(node.label),
                        properties=node.properties
                    )
                for i, node in enumerate(res)
            ]


class RedisSQErzoTransaction(CypherSQErzoTransaction):
    SUPPORTED_TYPES = ("str", "int", "float", "bool")

    def dump_data(self,
                  partial_query_nodes: dict,
                  partial_edged: dict):
        if partial_query_nodes:
            node_query = f"CREATE {', '.join(partial_query_nodes.values())}"
            self.graph.db_engine.query(node_query)

        #
        # Get al Edges
        #
        for edges in partial_edged.values():

            # Get first element for complete query
            edge = edges[0]

            #
            # Build bach information.
            #
            # We use indexed data batch, instead of dict, because not all
            # DB Engines support list of dict, but all supports indexed
            # data
            #
            batch = [
                [
                    b.source.make_identity(),
                    b.destination.make_identity(),
                    b.make_identity()
                ]
                for b in edges
            ]

            q = f"""
            UNWIND $batch as row
            MATCH (from:{edge.source.labels()} {{ identity: row[0] }})
            MATCH (to:{edge.destination.labels()} {{ identity: row[1] }})
            CREATE (from)-[:{edge.labels()} {{ identity: row[2] }}]->(to)
            """

            self.graph.db_engine.query(q, batch=batch)


class RedisSQErzoGraphConnection(CypherSQErzoGraphConnection):

    def __init__(self, connection_string: str):
        self.connection = self._parse_connection_string(
            connection_string
        )

    def update_element(self, graph_element: GraphElement) \
            -> None or SQErzoElementExistException:
        """Redis doesn't support multiple labels -> we'll simulate then"""

        def update_fixed_label_nodes(n) -> Iterable[Tuple[bool, GraphElement]]:
            if len(n.__labels__) > 1:
                #
                # If multiple labels are not supported, then create
                # net node clone with only new labels
                #
                for lb in n.__labels__[1:]:
                    new_node = n.clone(exclude=["identity"])
                    new_node.__labels__ = [lb]

                    yield True, new_node

            else:
                yield False, n

        if not graph_element:
            raise SQErzoElementExistException("Node update needs a valid Node")

        for need_create, n in update_fixed_label_nodes(graph_element):

            # No multiple labels supported. Need to be created a new node
            if need_create:
                q = n.query_create()

            # Multiple labels supported -> update node
            else:
                q = n.query_update()

            self.query(q)

            n.__dirty_properties__.clear()

    def save_element(self, graph_element: GraphElement) \
            -> None or SQErzoElementExistException:

        node_id = graph_element.make_identity()

        if isinstance(graph_element, GraphNode):
            if self.get_node_by_id(node_id):
                raise SQErzoElementExistException(
                    f"Graph element with id '{node_id}' already exits"
                )

        super(RedisSQErzoGraphConnection, self).save_element(graph_element)

    @property
    def transaction_class(self) -> Type:
        return RedisSQErzoTransaction

    def query(self, query: str, **kwargs):
        self.connection.query(query, kwargs)

    def query_response(self, query: str, **kwargs) -> Iterable[ResultElement]:
        return RedisGraphSQErzoQueryResponse(self, query, **kwargs)

    def create_constraints_nodes(self, key: str, labels: List[str]):
        """Redis doesn't support constraints -> do nothing"""

    def create_constraints_edges(self, key: str, labels: List[str]):
        """Redis doesn't support constraints -> do nothing"""

    def create_indexes(self, attribute: str, labels: List[str]):
        #
        # Redis doesn't support multi-label nodes -> create two index. One
        # for each node
        #
        for label in labels:
            q = f"""
            CREATE INDEX ON :{label}({attribute})
            """

            self.query(q)


    def _parse_connection_string(self, cs: str) -> Graph:
        parsed = pr.urlparse(cs)

        if parsed.path:
            db = parsed.path[1:]

            if not db:
                db = 0
            else:
                db = int(db)
        else:
            db = 0

        if parsed.query:
            qp = parsed.query.split("&", maxsplit=1)[0]
            p_name, p_value = qp.split("=")
            if p_name != "graph":
                db_graph = "sqerzo"
            else:
                db_graph = p_value
        else:
            db_graph = "sqerzo"

        port = parsed.port
        if port is None:
            port = 6379

        r = redis.Redis(
            host=parsed.hostname,
            port=port,
            username=parsed.username,
            password=parsed.password,
            db=db
        )

        return Graph(db_graph, r)

__all__ = ("RedisSQErzoGraphConnection",)



import uuid
import random

from collections import Iterable


def get_class_properties(c) -> Iterable:
    props = set()
    for m in c.mro():
        if m.__dict__.get("__annotations__", None):
            props.update(list(m.__dict__["__annotations__"].keys()))

    return props


guuid = lambda: uuid.uuid4().hex
rtext = lambda: "".join(random.choice(string.ascii_letters) for _ in range(80))


__all__ = ("get_class_properties", "guuid", "rtext")



from __future__ import annotations

import logging
import urllib.parse as pr

from typing import Type, List, Iterable

from .cache import *
from .config import *
from .exceptions import *
from .graph.model import *
from .graph.query import Query
from .graph.transaction import SQErzoTransaction
from .graph.interfaces import SQErzoGraphConnection
from .graph.cypher.neo4j import Neo4JSQErzoGraphConnection
from .graph.cypher.redisgraph import RedisSQErzoGraphConnection

log = logging.getLogger("sqerzo")

_SUPPORTED_TYPES = ("str",)
_SUPPORTED_TYPES_REDIS = ("str", "int", "float", "bool")
_SUPPORTED_TYPES_NEO4J = ("str", "int", "float", "bool", "datetime")


class SQErzoGraph:

    def __init__(self, connection_string: str = None, cache: str = None):
        self.connection_string = self._fix_connection_string(connection_string)
        self.db_engine: SQErzoGraphConnection = self._setup_db(self.connection_string)
        self.cache = self._setup_cache(cache)

        self._create_constrains()

    def transaction(self) -> SQErzoTransaction:
        return self.db_engine.transaction_class(self)

    def update(self, node: GraphElement):
        self.db_engine.update_element(node)

    def get_node_by_id(self, node_id: str, map_class: Type[GraphNode]) \
            -> Type or None:
        if not node_id:
            return

        # Check if is in cache
        if node_cache := self.cache.get_id(node_id):
            return node_cache

        if found := self.db_engine.get_node_by_id(node_id):
            node = map_class.from_query_results(found)

            self.cache.save_id(node_id, node)

            return node

    def save(self, graph_element: GraphElement) \
            -> None or SQErzoElementExistException:
        self.db_engine.save_element(graph_element)


    def truncate(self):
        """Remove all nodes and edges of database"""
        self.db_engine.truncate()

    @property
    def Q(self) -> Query:
        return self.db_engine.query_builder(self)

    @property
    def Query(self) -> Query:
        return self.db_engine.query_builder(self)

    def fetch_many(self,
                   node_type: Type[GraphElement] or GraphElement,
                   **kwargs) \
            -> Iterable[GraphElement] or SQErzoException:
        for x in self.db_engine.fetch_nodes(node_type, **kwargs):
            yield x

    def fetch_one(self,
                   node_type: Type[GraphElement] or GraphElement,
                   **kwargs) \
            -> Iterable[GraphElement] or SQErzoException:

        for res in self.fetch_many(node_type, **kwargs):
            return res

    # -------------------------------------------------------------------------
    # Private methods
    # -------------------------------------------------------------------------
    def _setup_cache(self, cache_config: str):
        if cache_config is None or cache_config == "memory://":
            return MemoryGraphNodeCache()
        else:
            raise ValueError(f"Invalid cache config: '{cache_config}'")

    def _fix_connection_string(self, cs: str) -> str:

        parsed = pr.urlparse(cs)

        if parsed.path == cs:
            return f"{cs}://"
        else:
            return cs

    def _setup_db(self, connection: str) -> SQErzoGraphConnection:

        parsed = pr.urlparse(connection)

        # -------------------------------------------------------------------------
        # Update connection_string
        # -------------------------------------------------------------------------
        if parsed.scheme.startswith("redis"):
            SQErzoConfig.SUPPORTED_TYPES = _SUPPORTED_TYPES_REDIS
            SQErzoConfig.SUPPORTED_INDEXES = True
            SQErzoConfig.SUPPORTED_CONSTRAINTS = False
            SQErzoConfig.SUPPORTED_MULTIPLE_LABELS = False

            return RedisSQErzoGraphConnection(self.connection_string)

        elif parsed.scheme.startswith(("enterprise+neo4j", "neo4j")):
            SQErzoConfig.SUPPORTED_TYPES = _SUPPORTED_TYPES_NEO4J
            SQErzoConfig.SUPPORTED_INDEXES = True
            SQErzoConfig.SUPPORTED_CONSTRAINTS = True
            SQErzoConfig.SUPPORTED_MULTIPLE_LABELS = True

            return Neo4JSQErzoGraphConnection(self.connection_string)

        elif parsed.scheme.startswith("neptune"):
            raise NotImplementedError("Neptune is not implemented yet")

        elif parsed.scheme.startswith("gremlin"):
            raise NotImplementedError("Neptune is not implemented yet")

        else:
            raise ValueError("Invalid db engine")

    def _create_constrains(self):

        nodes: List[GraphNode] = SQErzoConfig.SETUP_OBJECTS

        for n in nodes:
            labels = n.__labels__

            #
            # For Edge types
            #
            if issubclass(n, GraphEdge):

                for key in n.__keys__:
                    self.db_engine.create_constraints_edges(key, labels)

                for key in n.__unique__:
                    self.db_engine.create_constraints_nodes(key, labels)

            elif issubclass(n, GraphNode):

                for key in n.__keys__:
                    self.db_engine.create_constraints_nodes(key, labels)

                for key in n.__unique__:
                    self.db_engine.create_constraints_nodes(key, labels)

                for index_attr in n.__indexes__:
                    self.db_engine.create_indexes(index_attr, labels)

            else:
                raise ValueError(f"Invalid Node type: {type(n)}")

__all__ = ("SQErzoGraph",)



import abc

from .model import GraphElement


class SQErzoTransaction:

    @abc.abstractmethod
    def add(self, element: GraphElement):
        raise NotImplementedError()

    @abc.abstractmethod
    def dump_data(self, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def commit(self):
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.commit()


__all__ = ("SQErzoTransaction",)




from __future__ import annotations

import abc

from typing import List, Iterable, Type
from dataclasses import dataclass

from .model import GraphElement
from .query import Query
from ..exceptions import SQErzoElementExistException, SQErzoException


@dataclass
class ResultElement:
    id: str = None  # DB Id
    alias: str = None  # DB Id
    labels: List[str] = None
    properties: dict = None

class SQErzoQueryResponse:

    @abc.abstractmethod
    def __init__(self, graph: SQErzoGraphConnection, query: str, **kwargs):
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @abc.abstractmethod
    def __iter__(self):
        raise NotImplementedError()


class SQErzoGraphConnection(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __init__(self, connection_string: str):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_node_by_id(self, node_id: str) \
            -> GraphElement or None:
        raise NotImplementedError()

    @abc.abstractmethod
    def save_element(self, graph_element: GraphElement) \
            -> None or SQErzoElementExistException:
        raise NotImplementedError()

    @abc.abstractmethod
    def update_element(self, graph_element: GraphElement) \
            -> None or SQErzoElementExistException:
        raise NotImplementedError()

    @property
    def batch_size(self) -> int:
        return 1000

    @property
    @abc.abstractmethod
    def transaction_class(self) -> Type:
        raise NotImplementedError()

    @abc.abstractmethod
    def query(self, query: str, **kwargs) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def query_response(self, query: str, **kwargs) -> Iterable[ResultElement]:
        raise NotImplementedError()

    @abc.abstractmethod
    def fetch_nodes(self,
                    node_type: Type[GraphElement] or GraphElement,
                    **kwargs) \
            -> Iterable[GraphElement] or SQErzoException:
        raise NotImplementedError()

    @abc.abstractmethod
    def create_indexes(self, attribute: str, labels: List[str]):
        raise NotImplementedError()

    @abc.abstractmethod
    def create_constraints_nodes(self, key: str, labels: List[str]):
        raise NotImplementedError()

    @abc.abstractmethod
    def create_constraints_edges(self, key: str, labels: List[str]):
        raise NotImplementedError()

    @abc.abstractmethod
    def truncate(self):
        raise NotImplementedError()

    def query_builder(self, graph) -> Query:
        return Query(graph)

__all__ = ("ResultElement", "SQErzoGraphConnection",)




# -------------------------------------------------------------------------
# Utils
# -------------------------------------------------------------------------
from typing import List

from ...config import SQErzoConfig

def scape_string(text: str):
    return text.replace("""\\""", """\\\\""").replace("'", """\\'""")

def prepare_params(values: dict, operation="insert", node_name="a") -> List[str]:
    """Operation values: [insert|query|update]"""

    if not values:
        return []

    if operation == "insert":
        op = ":"
        nn = ""
    else:
        nn = f"{node_name}."
        op = ":"

    ret = []
    ret_append = ret.append

    for k, v in values.items():
        type_v = v.__class__.__name__

        if type_v == "str" and type_v in SQErzoConfig.SUPPORTED_TYPES:
            o = f"{nn}{k}{op} '{scape_string(v)}'"

        elif type_v == "datetime" and type_v in SQErzoConfig.SUPPORTED_TYPES:
            tm = v.strftime("%Y-%m-%dT%H:%M:%S%z")
            o = f"{nn}{k}{op} datetime('{tm}')"

        elif type_v == "bool" and type_v in SQErzoConfig.SUPPORTED_TYPES:
            o = f"{nn}{k}{op} {'true' if v else 'false'}"

        elif type_v == ("int", "float") and type_v in SQErzoConfig.SUPPORTED_TYPES:
            o = f"{nn}{k}{op} {v}"
        else:
            o = f"{nn}{k}{op} '{scape_string(str(v))}'"

        ret_append(o)

    return ret



def create_query(node, partial: bool = False):

    if not node.identity:
        node.identity = node.make_identity()

        # Do not include in dirty properties
        del node.__dirty_properties__["identity"]

    labels = node.labels()

    tmp_prop = [f"identity: '{node.identity}'"]
    tmp_prop.extend(prepare_params(node.properties))
    tmp_prop.extend(prepare_params({
        k: v
        for k, v in node.__dict__.items()
        if
        k not in ("properties", "identity") and not k.startswith("_")
    }))

    prop = f"{{{', '.join(tmp_prop)}}}"

    return f"{'' if partial else 'CREATE '} (:{labels} {prop})"

__all__ = ("create_query", )



from enum import IntEnum
from typing import List, Type

class _SQErzoConfig(dict):
    __slots__ = ()
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__

    def __init__(self, *args, **kwargs):
        super(_SQErzoConfig, self).__init__(*args, **kwargs)

        self["DB_ENGINE"]: DBEngine = None
        self["SUPPORTED_CONSTRAINTS"]: bool = False
        self["SUPPORTED_INDEXES"]: bool = False
        self["SUPPORTED_MULTIPLE_LABELS"]: bool = False
        self["SUPPORTED_TYPES"]: List[str] = []
        self["SETUP_OBJECTS"]: List[Type] = []

SQErzoConfig = _SQErzoConfig()

__all__ = ("SQErzoConfig",)



class MemoryGraphNodeCache:

    def __init__(self):
        self._cache = {}

    def make_key(self, node):

        if k := node.identity:
            return k
        else:
            k = "#".join(
                getattr(node, k)
                for k in node.__keys__
            )
            return k

    def get_id(self, node_id: str) -> object or None:
        try:
            return self._cache[node_id]
        except KeyError:
            return None

    def save_id(self, node_id: str, obj: object) -> object or None:
        self._cache[node_id] = obj

    def get(self, node: object) -> object or None:
        try:
            return self._cache[node.make_identity()]
        except KeyError:
            return None

    def save(self, node: object):
        key = self.make_key(node)
        self._cache[key] = node


__all__ = ("MemoryGraphNodeCache",)



from collections import defaultdict

from .lang import create_query
from ..transaction import SQErzoTransaction
from ..model import GraphElement, GraphNode, GraphEdge

class CypherSQErzoTransaction(SQErzoTransaction):

    def __init__(self, graph):
        self.graph = graph
        self._commit = []

    def add(self, element: GraphElement):
        self._commit.append(element)


    def dump_data(self,
                  partial_query_nodes: dict,
                  partial_edged: dict):

        if partial_query_nodes:
            node_query = f"CREATE {', '.join(partial_query_nodes.values())}"
            self.graph.db_engine.query(node_query)

        #
        # Get al Edges
        #
        for edges in partial_edged.values():

            # Get first element for complete query
            edge = edges[0]

            #
            # Build bach information.
            #
            batch = [
                {
                    "identity_from": b.source.make_identity(),
                    "identity_to": b.destination.make_identity(),
                    "identity_edge": b.make_identity()
                }
                for b in edges
            ]

            q = f"""
            UNWIND $batch as row
            MATCH (from:{edge.source.labels()} {{ identity: row.identity_from }})
            MATCH (to:{edge.destination.labels()} {{ identity: row.identity_to }})
            CREATE (from)-[:{edge.labels()} {{ identity: row.identity_edge }}]->(to)
            """

            self.graph.db_engine.query(q, batch=batch)


    def commit(self):

        counter = 0
        processed = 0
        partial_query_nodes = {}
        partial_query_nodes_processed = set()
        partial_edges = defaultdict(list)
        chunk_size = self.graph.db_engine.batch_size

        #
        # Get all nodes and create the query
        #
        # Use this approach instead of a function that split list in chunks
        # to avoid the memory of the chunked list, the creation of
        # a new list and scroll through the list 2 times. For big commit
        # list it should be a problem.
        #
        while self._commit:

            element = self._commit.pop()

            if isinstance(element, GraphNode):
                key = "".join(getattr(element, x) for x in element.__keys__)

                if key in partial_query_nodes_processed:
                    continue

                partial_query_nodes_processed.add(key)
                partial_query_nodes[key] = create_query(element, partial=True)

            elif isinstance(element, GraphEdge):
                label_edge = element.labels()
                label_source = element.source.labels()
                label_destination = element.destination.labels()
                key = f"{label_source}#{label_edge}#{label_destination}"

                partial_edges[key].append(element)

            counter += 1
            processed += 1

            if counter >= chunk_size:
                self.dump_data(partial_query_nodes, partial_edges)

                #
                # Reset
                #
                counter = 0
                partial_edges.clear()  # Reset list
                partial_query_nodes.clear()  # Reset list

        # Check if there are some remain data to dump
        self.dump_data(partial_query_nodes, partial_edges)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.commit()

__all__ = ("CypherSQErzoTransaction",)



from typing import Dict, Type

from ..exceptions import SQErzoException

class Query:

    def __init__(self, graph):
        self.graph = graph
        self.query = None

    def raw(self, query: str = None):
        self.query = query

        return self

    def execute(self, map_to: Dict[str, Type] = None):
        if self.query:
            with self.graph.db_engine.query_response(self.query) as responses:
                results = []

                for row in responses:
                    tmp = []

                    if map_to:
                        for column in row:

                            try:
                                tmp.append(
                                    map_to[column.alias].from_query_results(column)
                                )
                            except KeyError:
                                raise SQErzoException(
                                    f"Can't find class for mapping alias '{row.alias}'"
                                )

                        results.append(tmp)

                    else:
                        results.append(
                            row
                        )

                return results



from __future__ import annotations

import abc
import hashlib
import logging

from typing import List
from collections import Iterable
from dataclasses import dataclass, field

from ..exceptions import SQErzoException
from .helpers import get_class_properties, guuid

log = logging.getLogger("sqerzo")

class DirtyDict(dict):
    __slots__ = ()
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__

    def __init__(self, *args, **kwargs):
        super(DirtyDict, self).__init__(*args, **kwargs)

        self.__dirty_properties__ = {}

    def __setitem__(self, key, value):
        try:
            old_value = self.__dict__[key]
        except KeyError:
            old_value = None

        self.__dirty_properties__[key] = old_value

        super(DirtyDict, self).__setitem__(key, value)


    def items(self):
        for k, v in super(DirtyDict, self).items():
            if k.startswith("_"):
                continue

            yield k, v

class GraphElementMetaClass(type):
    """
    This is the metaclass for each GraphNode or GraphEdge subclass. It checks
    that all nodes have these properties when created:

    - __keys__
    - __labels__
    - __unique__
    - __indexes__

    Something important things:

    **__labels__**

    If not property __label__ was set into class, it will try to deduce from
    class name. Valid class names for deducing must ends with 'Node' word. For
    example: MyCustomNode

    **__keys__**

    If not __keys__ property was set the default value are: 'identity' property

    **__unique__**

    It's an iterable value.

    **__indexes__**

    It's an iterable value.
    """

    def __new__(cls, class_name: str, bases, dct: dict):

        o = type.__new__(cls, class_name, bases, dct)

        if class_name in ("GraphElement", "GraphNode", "GraphEdge"):
            return o

        iterable_types = (list, set, tuple)

        #
        # Ensures that all nodes has the property: __label__
        #
        if not hasattr(o, "__labels__"):
            #
            # Try to find label by class name
            #
            log.debug(f"Trying to deduce Node label from class name: "
                      f"'{class_name}'")

            if (index := class_name.rfind("Node")) != -1:
                label = class_name[:index]

            elif (index := class_name.rfind("Edge")) != -1:
                label = class_name[:index]
            else:
                label = None

            if not label:
                raise SQErzoException(
                    f"Class name '{class_name}' must has property __label__ or "
                    f"the class name must ends with 'Node' or 'Edge' name. "
                    f"Examples: MyCustomNode, MyCustomEdge "
                )
            else:
                o.__labels__ = {label}

        else:
            # Check types
            if not isinstance(o.__labels__, iterable_types):
                o.__labels__ = [o.__labels__]

            o.__labels__ = set(o.__labels__)

        try:
            class_props = get_class_properties(o)
        except KeyError:
            raise SQErzoException(
                "Currently only support Python Data Classes"
            )

        if not hasattr(o, "__keys__"):
            o.__keys__ = {"identity"}

        else:
            # Check types
            if not isinstance(o.__keys__, iterable_types):
                o.__keys__ = [o.__keys__]

            o.__keys__ = set(o.__keys__)

            #
            # Check attribute values are in class properties
            #
            if missing := o.__keys__ - class_props:
                raise SQErzoException(
                    f"__keys__ has a value missing in class property "
                    f"'{class_name}': {missing} "
                )


        if hasattr(o, "__unique__"):

            # Check types
            if not isinstance(o.__unique__, iterable_types):
                o.__unique__ = [o.__unique__]

            # Fix duplicates
            o.__unique__ = set(o.__unique__)

            #
            # Check attribute values are in class properties
            #
            if missing := o.__unique__ - class_props:
                raise SQErzoException(
                    f"__unique__ has a value missing in class property "
                    f"'{class_name}': {missing} "
                )
        else:
            o.__unique__ = set()

        if hasattr(o, "__indexes__"):

            # Check types
            if not isinstance(o.__indexes__, iterable_types):
                o.__indexes__ = [o.__indexes__]

            # Fix duplicates
            o.__indexes__ = set(o.__indexes__)
            o.__indexes__.add("identity")  # Ensures identity

            #
            # Check attribute values are in class properties
            #
            if missing := o.__indexes__ - class_props:
                raise SQErzoException(
                    f"__indexes__ has a value missing in class property "
                    f"'{class_name}': {missing} "
                )
        else:
            o.__indexes__ = {"identity"}

        #
        # Track object
        #
        from sqerzo.config import SQErzoConfig

        SQErzoConfig.SETUP_OBJECTS.append(o)

        # Attack meta properties
        o.__dirty_properties__ = {}
        o.__is_instance__ = False

        return o

    def __call__(cls, *args, **kwargs):
        o = super().__call__(*args, **kwargs)
        o.__is_instance__ = True

        return o

#
# Base classes
#
class GraphElement(metaclass=GraphElementMetaClass):

    def __setattr__(self, key, value):
        if self.__is_instance__ and key != "__is_instance__":

            # Save old value
            self.__dirty_properties__[key] = getattr(self, key, None)

        super(GraphElement, self).__setattr__(key, value)

    def add_label(self, label_name: str):
        if not label_name:
            return

        self.__labels__.add(label_name)

    @abc.abstractmethod
    def clone(self, exclude: List[str] = None) -> GraphElement:
        raise NotImplementedError()

    @abc.abstractmethod
    def query_create(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def query_update(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def query_get_node(self) -> str:
        raise NotImplementedError()

    def labels(self) -> str:
        return ":".join(self.__labels__)


@dataclass
class GraphNode(GraphElement):
    properties: DirtyDict = field(default_factory=DirtyDict)
    identity: str = None

    @classmethod
    def from_query_results(cls, result_data: object or list[object]):
        properties = {
            k:v for k, v in result_data.properties.items()
            if k not in ("identifier", "alias")
        }
        custom_class_properties = {
            k: result_data.properties[k]
            for k in cls.__annotations__.keys()
        }

        config = {
            **custom_class_properties,
            "identity": result_data.properties["identity"],
            "properties": DirtyDict(properties)
        }

        return cls(**config)

    def make_identity(self) -> str:
        if self.identity:
            return self.identity

        l = self.labels()

        key_builders = [*self.__keys__]

        if not (set(self.__keys__) - {"identity"}) and not self.identity:
            self.identity = guuid()

        q = "#".join(
            f"{k}:{'' if v is None else v}"
            for k, v in self.__dict__.items()
            if k in key_builders
        )

        v =  hashlib.sha512(f"{l}#{q}".encode()).hexdigest()

        self.identity = v

        return v

@dataclass
class GraphEdge(GraphElement):
    source: GraphNode
    destination: GraphNode
    properties: dict = field(default_factory=dict)
    identity: str = None

    def make_identity(self) -> str:
        if self.identity:
            return self.identity

        l = self.labels()
        n = f"{self.source.make_identity()}#{self.source.make_identity()}"

        v =  hashlib.sha512(f"{l}#{n}".encode()).hexdigest()

        self.identity = v

        return v


__all__ = ("GraphNode", "GraphEdge", "GraphElement", "DirtyDict")
